import _ from "lodash";
import { ApolloClient } from "apollo-client";
import { ApolloLink } from "apollo-link";
import { getMainDefinition } from "apollo-utilities";
import isBrowser from "./isBrowser";
import Authenticator from "./Authenticator";
import ErrorLink from "./links/ErrorLink";
import RetryLink from "./links/RetryLink";
import AuthLink from "./links/AuthLink";
import HttpLink from "./links/HttpLink";
import WebSocketLink from "./links/WebSocketLink";
import ApolloListener from "./ApolloListener";

function createLink(configs, auth) {
  const { uri } = configs;

  const httpLink = new HttpLink({ uri, ...configs.http }, auth);

  if (!isBrowser) return httpLink;

  const wsLink = new WebSocketLink({ uri, ...configs.ws }, auth);

  return ApolloLink.split(
    ({ query }) => {
      const { kind, operation } = getMainDefinition(query);
      return kind === "OperationDefinition" && operation === "subscription";
    },
    wsLink,
    httpLink,
  );
}

function get(target, name) {
  if (typeof Authenticator.prototype[name] === "function") {
    return target.auth[name].bind(target.auth);
  }
  return target[name];
}

export default class extends ApolloClient {
  constructor(options) {
    const { cache, connectToDevTools, ...configs } = _.defaults({}, options, {
      uri: "http://localhost:3000/",
      onError: () => {},
    });
    const { uri } = configs;

    const auth = new Authenticator({ uri, ...configs.auth });

    super({
      connectToDevTools,
      ssrMode: !isBrowser,
      link: ApolloLink.from([
        new ErrorLink(configs.onError, auth),
        new AuthLink(configs.auth, auth),
        new RetryLink(configs.retry, auth),
        createLink(configs, auth),
      ]),
      cache,
    });

    this.auth = auth;
    this.configs = configs;

    const resetStore = () => this.resetStore();
    ApolloListener.on("signInResolve", resetStore);
    ApolloListener.on("signOutResolve", resetStore);
    ApolloListener.on("changeViewerResolve", resetStore);

    return new Proxy(this, { get });
  }
}
