import _ from "lodash";
import axios from "axios";
import URL from "url-parse";
import { assertResult } from "thelper";
import ApolloListener from "./ApolloListener";
import ApolloToken from "./ApolloToken";

function isForbidden(e) {
  return _.get(e, ["response", "status"]) === 403;
}

const events = Object.freeze({
  signInResolve: ["signInResolve", "signIn", "resolve"],
  signInReject: ["signInReject", "signIn", "reject"],
  signOutResolve: ["signOutResolve", "signOut", "resolve"],
  signOutReject: ["signOutReject", "signOut", "reject"],
  changeViewerResolve: ["changeViewerResolve", "changeViewer", "resolve"],
  changeViewerReject: ["changeViewerReject", "changeViewer", "reject"],
  renewResolve: ["renewResolve", "renew", "resolve"],
  renewReject: ["renewReject", "renew", "reject"],
});

export default class Authenticator {
  instance = null;

  constructor(options) {
    const { uri, headers, ...configs } = _.defaults({}, options, {
      path: "/auth",
      params: {},
    });

    const url = new URL(uri);

    this.configs = configs;
    this.instance = axios.create({
      baseURL: url.origin,
      headers,
      withCredentials: true,
    });
  }

  async request(params, retry = 3) {
    try {
      const { instance, configs } = this;
      const result = await instance.post(configs.path, {
        ...configs.params,
        ...params,
      });
      return result.data;
    } catch (e) {
      if (isForbidden(e) || retry < 1) throw e;

      await new Promise(r => setTimeout(r, _.random(1500, 500)));
      return this.request(params, retry - 1);
    }
  }

  async signIn(params) {
    try {
      const result = await this.request({ method: "signIn", ...params });
      if (result.isUnfamiliar) return result;
      ApolloToken.apply(result);
      ApolloListener.emit(events.signInResolve, result);
      return result;
    } catch (error) {
      ApolloListener.emit(events.signInReject, { error });
      throw error;
    }
  }

  async signOut(params) {
    try {
      const result = await this.request({ method: "signOut", ...params });
      ApolloListener.emit(events.signOutResolve, result);
      return result;
    } catch (error) {
      ApolloListener.emit(events.signOutReject, { error });
      throw error;
    } finally {
      ApolloToken.apply({ accessToken: "", refreshToken: "" });
    }
  }

  async changeViewer(...args) {
    try {
      const result = await this.renew(...args);
      ApolloToken.apply(result);
      ApolloListener.emit(events.changeViewerResolve, result);
      return result;
    } catch (error) {
      ApolloListener.emit(events.changeViewerReject, { error });
      throw error;
    }
  }

  async renew(params) {
    const { refreshToken } = ApolloToken;
    assertResult(refreshToken, new Error("refreshToken is null"));

    try {
      const result = await this.request({
        method: "renew",
        refreshToken,
        ...params,
      });
      ApolloToken.apply(result);
      ApolloListener.emit(events.renewResolve, result);
      return result;
    } catch (error) {
      ApolloToken.apply({ accessToken: "", refreshToken: "" });
      ApolloListener.emit(events.renewReject, { error });
      throw error;
    }
  }
}
