import "isomorphic-unfetch";
import gql from "graphql-tag";
import { AuthenticationError } from "apollo-server-express";
import localStorage from "../localStorage";
import { initApollo } from "..";
import server, { jwt, context, onRenew } from "../__mocks__/server";

const auth = {
  headers: { cookie: "x-correlation-id=XXYYZZ;" },
  params: { origin: "http://host" },
};

describe("initApollo with http", () => {
  let uri;

  beforeAll(async () => {
    uri = await server.listen();
  });

  afterAll(async () => server.close());

  beforeEach(() => {
    localStorage.clear();
  });

  const request = options => {
    const client = initApollo({}, { uri, auth, ...options });
    return client.query({
      query: gql`
        query {
          viewer
        }
      `,
      fetchPolicy: "network-only",
    });
  };

  it("successfully get data", async () => {
    const { data } = await request();
    expect(data).toEqual({ viewer: 1 });
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-refresh-token");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-access-token");
    expect(localStorage.getItem).toHaveBeenCalledTimes(2);
    expect(localStorage.setItem).toHaveBeenCalledTimes(0);
    expect(context).toHaveBeenCalledWith(
      expect.objectContaining({ authorization: "" }),
    );
  });

  it("successfully get data with token", async () => {
    localStorage.apply({
      "apollo-refresh-token": "REFRESH_TOKEN",
      "apollo-access-token": "ACCESS_TOKEN",
      "apollo-expired": Date.now() + 60 * 1000,
    });
    const { data } = await request();
    expect(data).toEqual({ viewer: 1 });
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-refresh-token");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-expired");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-access-token");
    expect(localStorage.getItem).toHaveBeenCalledTimes(3);
    expect(localStorage.setItem).toHaveBeenCalledTimes(0);
    expect(context).toHaveBeenCalledWith(
      expect.objectContaining({ authorization: "Bearer ACCESS_TOKEN" }),
    );
  });

  describe("when AuthenticationError", () => {
    const expectToReject = async () => {
      let err;
      try {
        await request();
      } catch (e) {
        err = e;
      }
      expect(err.networkError.result.errors[0].extensions).toEqual({
        code: "UNAUTHENTICATED",
      });
    };

    it("successfully renew access token", async () => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({
        "apollo-refresh-token": token,
        "apollo-extend": Date.now() - 60 * 1000,
      });
      const { data } = await request();
      expect(data).toEqual({ viewer: 1 });
      expect(onRenew).toHaveBeenCalledWith({
        method: "renew",
        correlationId: "XXYYZZ",
        issuer: "private",
        origin: "http://host",
        id: "10",
      });
      expect(onRenew).toHaveBeenCalledTimes(1);
      const expired = localStorage.getItem("apollo-expired");
      expect(expired).toBeGreaterThan(Date.now());
      const accessToken = localStorage.getItem("apollo-access-token");
      expect(accessToken).not.toEqual(token);
      const refreshToken = localStorage.getItem("apollo-refresh-token");
      expect(refreshToken).not.toBeUndefined();
      expect(localStorage.setItem).toHaveBeenCalledTimes(3);
    });

    it("with renew failed", async () => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({
        "apollo-refresh-token": token,
        "apollo-extend": Date.now() - 60 * 1000,
      });
      context.mockRejectedValueOnce(new AuthenticationError());
      onRenew.mockRejectedValueOnce(new Error("blockade"));
      await expectToReject();
      expect(onRenew).toHaveBeenCalledTimes(1);
    });
  });

  describe("when server error", () => {
    it("successfully retry", async () => {
      context.mockImplementationOnce(() => {
        throw new Error();
      });
      const onError = jest.fn();
      const { data } = await request({ onError });
      expect(data).toEqual({ viewer: 1 });
      expect(context).toHaveBeenCalledTimes(2);
      expect(onError).toHaveBeenCalledTimes(0);
    });

    it("with failed after retry", async () => {
      context.mockImplementationOnce(() => {
        throw new Error();
      });
      const onError = jest.fn();
      const error = new Error(
        "Network error: Response not successful: Received status code 500",
      );
      await expect(request({ onError, retry: { max: 1 } })).rejects.toEqual(
        error,
      );
      expect(context).toHaveBeenCalledTimes(1);
      expect(onError).toHaveBeenCalledTimes(1);
    });
  });
});
