import "isomorphic-unfetch";
import localStorage from "../localStorage";
import { initApollo, ApolloListener } from "..";
import server, { jwt, onSignIn, onSignOut, onRenew } from "../__mocks__/server";

const auth = {
  headers: { cookie: "x-correlation-id=XXYYZZ;" },
  params: { origin: "http://host" },
};

const waitPubSub = () => new Promise(r => setTimeout(r, 10));

describe("signIn & signOut", () => {
  let uri;
  let client;

  beforeAll(async () => {
    uri = await server.listen();
  });

  afterAll(async () => server.close());

  beforeEach(() => {
    localStorage.clear();
    client = initApollo({}, { uri, auth });
  });

  describe("signIn", () => {
    it("successfully got user and token", async () => {
      onSignIn.mockReturnValueOnce({
        correlationId: "NEW",
        id: "10",
      });
      const resolve = jest.fn();
      const listener = ApolloListener.on("signInResolve", resolve);

      const body = { username: "apple", password: "applepen" };
      const result = await client.signIn(body);
      expect(result.id).toEqual("10");
      expect(localStorage.setItem).toHaveBeenCalledWith(
        "apollo-access-token",
        expect.any(String),
      );
      expect(localStorage.setItem).toHaveBeenCalledWith(
        "apollo-refresh-token",
        expect.any(String),
      );
      expect(onSignIn).toHaveBeenCalledWith({
        method: "signIn",
        username: "apple",
        password: "applepen",
        correlationId: "XXYYZZ",
        origin: "http://host",
      });
      expect(onSignIn).toHaveBeenCalledTimes(1);
      await waitPubSub();
      expect(resolve).toHaveBeenCalledWith(
        expect.objectContaining({
          accessToken: expect.any(String),
          refreshToken: expect.any(String),
        }),
      );
      expect(resolve).toHaveBeenCalledTimes(1);
      listener();
    });

    it("when isUnfamiliar", async () => {
      onSignIn.mockReturnValueOnce({
        correlationId: "NEW",
        id: "10",
        isUnfamiliar: true,
      });
      const body = { username: "apple", password: "applepen" };
      const result = await client.signIn(body);
      expect(result.id).toBe("10");
      expect(localStorage.setItem).toHaveBeenCalledTimes(0);
    });

    it("when signIn failed", async () => {
      const reject = jest.fn();
      const listener = ApolloListener.on("signInReject", reject);

      onSignIn.mockRejectedValueOnce(new Error("not match password"));

      const error = new Error("Request failed with status code 403");
      await expect(client.signIn()).rejects.toEqual(error);
      await waitPubSub();
      expect(reject).toHaveBeenCalledWith(
        expect.objectContaining({ error: expect.any(Error) }),
      );
      expect(reject).toHaveBeenCalledTimes(1);
      listener();
    });
  });

  describe("signOut", () => {
    beforeEach(() => {
      localStorage.apply({
        "apollo-refresh-token": jwt.signRefresh("XXYYZZ", { id: "10" }),
        "apollo-access-token": jwt.signAccess({ id: "10" }),
        "apollo-expired": Date.now() + 50 * 60 * 1000,
      });
    });

    it("successfully signOut", async () => {
      const resolve = jest.fn();
      const listener = ApolloListener.on("signOutResolve", resolve);

      const body = { clear: true };
      const result = await client.signOut(body);
      expect(result).toEqual({ status: "ok" });
      expect(localStorage.getItem("apollo-access-token")).toBe("");
      expect(localStorage.getItem("apollo-refresh-token")).toBe("");
      expect(onSignOut).toHaveBeenCalledWith({
        method: "signOut",
        correlationId: "XXYYZZ",
        clear: true,
        origin: "http://host",
      });
      expect(onSignOut).toHaveBeenCalledTimes(1);
      await waitPubSub();
      expect(resolve).toHaveBeenCalledTimes(1);
      listener();
    });

    it("when singOut failed", async () => {
      const reject = jest.fn();
      const listener = ApolloListener.on("signOutReject", reject);
      onSignOut.mockRejectedValueOnce(new Error("system error"));

      const error = new Error("Request failed with status code 403");
      await expect(client.signOut()).rejects.toEqual(error);
      expect(localStorage.getItem("apollo-access-token")).toBe("");
      expect(localStorage.getItem("apollo-refresh-token")).toBe("");
      await new Promise(r => setTimeout(r, 200));
      expect(reject).toHaveBeenCalledWith(
        expect.objectContaining({ error: expect.any(Error) }),
      );
      expect(reject).toHaveBeenCalledTimes(1);
      listener();
    });
  });

  describe("renew", () => {
    beforeEach(() => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({ "apollo-refresh-token": token });
    });

    it("successfully renew", async () => {
      const resolve = jest.fn();
      const listener = ApolloListener.on("renewResolve", resolve);

      const body = { key: "XYZ" };
      const result = await client.renew(body);

      expect(result).toEqual(expect.objectContaining({ status: "ok" }));
      expect(onRenew).toHaveBeenCalledWith({
        method: "renew",
        correlationId: "XXYYZZ",
        issuer: "private",
        key: "XYZ",
        origin: "http://host",
        id: "10",
      });
      expect(onRenew).toHaveBeenCalledTimes(1);
      expect(resolve).toHaveBeenCalledWith(
        expect.objectContaining({
          accessToken: expect.any(String),
          refreshToken: expect.any(String),
        }),
      );
      expect(resolve).toHaveBeenCalledTimes(1);
      listener();
    });

    it("when renew failed", async () => {
      const reject = jest.fn();
      const listener = ApolloListener.on("renewReject", reject);
      onRenew.mockRejectedValueOnce(new Error("server reject"));

      const error = new Error("Request failed with status code 403");
      await expect(client.renew()).rejects.toEqual(error);
      await new Promise(r => setTimeout(r, 200));
      expect(localStorage.setItem).toHaveBeenCalledWith(
        "apollo-refresh-token",
        "",
      );
      expect(reject).toHaveBeenCalledWith(
        expect.objectContaining({ error: expect.any(Error) }),
      );
      expect(reject).toHaveBeenCalledTimes(1);
      listener();
    });
  });

  describe("changeViewer", () => {
    beforeEach(() => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({ "apollo-refresh-token": token });
    });

    it("successfully change viewer", async () => {
      const resolve = jest.fn();
      const listener = ApolloListener.on("changeViewerResolve", resolve);

      const body = { key: "XYZ" };
      const result = await client.changeViewer(body);

      expect(result).toEqual(expect.objectContaining({ status: "ok" }));
      expect(onRenew).toHaveBeenCalledWith({
        method: "renew",
        correlationId: "XXYYZZ",
        issuer: "private",
        key: "XYZ",
        origin: "http://host",
        id: "10",
      });
      expect(onRenew).toHaveBeenCalledTimes(1);
      expect(resolve).toHaveBeenCalledWith(
        expect.objectContaining({
          accessToken: expect.any(String),
          refreshToken: expect.any(String),
        }),
      );
      expect(resolve).toHaveBeenCalledTimes(1);
      listener();
    });

    it("when renew failed", async () => {
      const reject = jest.fn();
      const listener = ApolloListener.on("changeViewerReject", reject);
      onRenew.mockRejectedValueOnce(new Error("server reject"));

      const error = new Error("Request failed with status code 403");
      await expect(client.changeViewer()).rejects.toEqual(error);
      await new Promise(r => setTimeout(r, 200));
      expect(localStorage.setItem).toHaveBeenCalledWith(
        "apollo-refresh-token",
        "",
      );
      expect(reject).toHaveBeenCalledWith(
        expect.objectContaining({ error: expect.any(Error) }),
      );
      expect(reject).toHaveBeenCalledTimes(1);
      listener();
    });
  });
});
