/* eslint-disable import/first */

jest.mock("../isBrowser", () => ({ default: true }));

import "isomorphic-unfetch";
import ws from "ws";
import gql from "graphql-tag";
import { AuthenticationError } from "apollo-server-express";
import localStorage from "../localStorage";
import { initApollo } from "..";
import server, { jwt, pubsub, context, onRenew } from "../__mocks__/server";

const auth = { headers: { cookie: "x-correlation-id=XXYYZZ;" } };

describe("initApollo with ws", () => {
  let uri;

  beforeAll(async () => {
    uri = await server.listen();
  });

  afterAll(async () => server.close());

  beforeEach(() => {
    localStorage.clear();
  });

  const request = async options => {
    delete initApollo.client;

    const client = initApollo(
      {},
      { uri, ws: { webSocketImpl: ws }, auth, ...options },
    );

    let subscription;
    const result = await new Promise(async (subResolve, subReject) => {
      subscription = await pubsub.asyncIterator(["NEW_SUBSCRIPTION"]);
      client
        .subscribe({
          query: gql`
            subscription {
              viewer
            }
          `,
        })
        .subscribe({ next: subResolve, error: subReject });
      await subscription.next();
      pubsub.publish("NEW_VIEWER", { viewer: 1 });
    });
    await subscription.return();
    return result;
  };

  it("successfully get data", async () => {
    const { data } = await request();
    expect(data).toEqual({ viewer: 1 });
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-cache-persist");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-refresh-token");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-access-token");
    expect(localStorage.getItem).toHaveBeenCalledTimes(3);
    expect(localStorage.setItem).toHaveBeenCalledTimes(0);
    expect(context).toHaveBeenCalledWith({ authorization: "" });
    expect(context).toHaveBeenCalledTimes(1);
  });

  it("successfully get data with token", async () => {
    localStorage.apply({
      "apollo-refresh-token": "REFRESH_TOKEN",
      "apollo-access-token": "ACCESS_TOKEN",
      "apollo-expired": Date.now() + 60 * 1000,
    });
    const { data } = await request();
    expect(data).toEqual({ viewer: 1 });
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-cache-persist");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-refresh-token");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-expired");
    expect(localStorage.getItem).toHaveBeenCalledWith("apollo-access-token");
    expect(localStorage.getItem).toHaveBeenCalledTimes(4);
    expect(localStorage.setItem).toHaveBeenCalledTimes(0);
    expect(context).toHaveBeenCalledWith({
      authorization: "Bearer ACCESS_TOKEN",
    });
    expect(context).toHaveBeenCalledTimes(1);
  });

  it("successfully get client from memory", async () => {
    const options = { uri, ws: { webSocketImpl: ws } };

    const client = initApollo({}, options);
    expect(client).toBe(initApollo({}, options));
    await new Promise(res => setTimeout(res, 500));
    expect(context).toHaveBeenCalledTimes(0);
  });

  describe("when AuthenticationError", () => {
    const expectToReject = async () => {
      let err;
      try {
        await request();
      } catch (e) {
        err = e;
      }
      expect(err.message).toBe("token invalid");
    };

    it("successfully renew access token", async () => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({
        "apollo-refresh-token": token,
        "apollo-extend": Date.now() - 60 * 1000,
      });
      await request();
      expect(onRenew).toHaveBeenCalledTimes(1);
      const expired = localStorage.getItem("apollo-expired");
      expect(expired).toBeGreaterThan(Date.now());
      const accessToken = localStorage.getItem("apollo-access-token");
      expect(accessToken).not.toEqual(token);
      const refreshToken = localStorage.getItem("apollo-refresh-token");
      expect(refreshToken).not.toBeUndefined();
      expect(localStorage.setItem).toHaveBeenCalledTimes(3);
    });

    it("with renew failed", async () => {
      const token = jwt.signRefresh("XXYYZZ", { id: "10" });
      localStorage.apply({
        "apollo-refresh-token": token,
        "apollo-extend": Date.now() - 60 * 1000,
      });
      context.mockRejectedValueOnce(new AuthenticationError("token invalid"));
      onRenew.mockRejectedValueOnce(new Error("blockade"));
      await expectToReject();
      expect(onRenew).toHaveBeenCalledTimes(1);
    });
  });
});
