import _ from "lodash";
import crypto from "crypto";
import cookie from "cookie";
import { assertEnv, assertResult } from "thelper";
import { json } from "body-parser";
import JSONWebBase58 from "jw58";
import { AuthenticationError } from "apollo-server-errors";

const EXPIRED_IN = 14 * 24 * 60 * 60; // 7 days in seconds

function assertKeys(keys) {
  assertResult(keys, new TypeError("keys least one is required"));

  if (_.size(keys) > 1) return [keys];
  return _.map(keys, (rsa, name) => [rsa, name])[0];
}

export default class JsonWebToken {
  cryptor = null;

  constructor(keys, options) {
    this.cryptor = new JSONWebBase58(...assertKeys(keys));

    _.assign(
      this,
      _.defaults(options, {
        viewerName: "viewer",
        expiresIn: EXPIRED_IN,
        toUser: _.identity,
        toViewer: _.identity,
        onSignIn: _.identity,
        onSignOut: _.identity,
        onRenew: _.identity,
      }),
    );

    return this;
  }

  sign(...args) {
    return this.cryptor.sign(...args);
  }

  verify(...args) {
    return this.cryptor.verify(...args);
  }

  signRefresh(correlationId, passport, ...args) {
    return this.sign(
      {
        sub: "auth_refresh",
        exp: Math.floor(Date.now() / 1000) + this.expiresIn,
        correlationId,
        passport,
      },
      ...args,
    );
  }

  signAccess(passport, ...args) {
    return this.sign(
      {
        sub: "auth_access",
        exp: Math.floor(Date.now() / 1000) + 60 * 60,
        passport,
      },
      ...args,
    );
  }

  verifyRefresh(token, options) {
    const { correlationId, passport, iss } = this.verify(token, {
      sub: "auth_refresh",
      ...options,
    });
    return { ...passport, correlationId, issuer: iss };
  }

  verifyAccess(token, options) {
    if (!token) return {};
    const { passport, iss } = this.verify(token, {
      sub: "auth_access",
      ...options,
    });
    return {
      user: this.toUser(passport),
      viewer: this.toViewer(passport),
      issuer: iss,
    };
  }

  async signIn(params) {
    const result = await this.onSignIn(params);
    const { issuer, correlationId, isUnfamiliar, ...passport } = result;

    const id =
      result.correlationId ||
      params.correlationId ||
      crypto.randomBytes(16).toString("hex");

    _.forEach(passport, data => {
      assertResult(
        _.includes(["boolean", "string", "number"], typeof data) ||
          _.isPlainObject(data),
        new TypeError(
          "results must is boolean, string, number or plain object on sign-in",
        ),
      );
    });

    if (!isUnfamiliar) {
      const refreshToken = this.signRefresh(id, passport, issuer);
      const accessToken = this.signAccess(passport, issuer);
      return { ...passport, correlationId: id, refreshToken, accessToken };
    }

    return { ...passport, correlationId: id, isUnfamiliar };
  }

  async signOut(params) {
    assertResult(params.correlationId, new Error("correlationId is required"));
    return this.onSignOut(params);
  }

  async renew(params) {
    assertResult(params.refreshToken, new Error("refreshToken is required"));
    assertResult(params.correlationId, new Error("correlationId is required"));

    const signature = this.verifyRefresh(params.refreshToken);

    if (params.correlationId !== signature.correlationId) {
      throw new Error("not match to correlation id");
    }

    const { correlationId, ...passport } = await this.onRenew({
      ..._.omit(params, ["refreshToken"]),
      ...signature,
    });

    _.forEach(passport, data => {
      assertResult(
        _.includes(["boolean", "string", "number"], typeof data) ||
          _.isPlainObject(data),
        new TypeError(
          "results must is boolean, string, number or plain object on renew",
        ),
      );
    });

    const id = correlationId || signature.correlationId;
    const refreshToken = this.signRefresh(id, passport, signature.issuer);
    const accessToken = this.signAccess(passport, signature.issuer);
    return { ...passport, correlationId: id, refreshToken, accessToken };
  }

  applyMiddleware(app, options) {
    const { path } = _.defaults({}, options, { path: "/auth" });

    app.post(path, json());
    app.post(path, async (req, res) => {
      const { method } = req.body;
      if (!_.includes(["signIn", "signOut", "renew"], method)) {
        res.sendStatus(404);
        return;
      }

      let correlationId;
      try {
        correlationId = cookie.parse(req.headers.cookie)["x-correlation-id"];
      } catch (e) {
        // empty
      }

      try {
        const result = await this[method]({ ...req.body, correlationId });
        if (result.correlationId) {
          res.append(
            "Set-Cookie",
            cookie.serialize("x-correlation-id", result.correlationId, {
              httpOnly: true,
              maxAge: 365 * 24 * 60 * 60,
            }),
          );
        }
        res.json({ status: "ok", ..._.omit(result, ["correlationId"]) }).end();
      } catch (e) {
        res.status(403).send({ error: e.message });
      }
    });
  }

  contextParser() {
    const { viewerName } = this;
    return async context => {
      const { req, res, connection } = context;

      let token;
      try {
        const { authorization } = connection ? connection.context : req.headers;
        [, token] = /^Bearer (.+)$/.exec(authorization);
      } catch (e) {
        // empty
      }

      try {
        const { user, viewer, issuer } = await this.verifyAccess(token);
        return { ...context, user, [viewerName]: viewer, issuer };
      } catch (e) {
        assertEnv(() => {
          if (!res) throw new AuthenticationError(e.message);
          res.append("X-Content-Extend", e.message);
        });
        throw new AuthenticationError("token invalid");
      }
    };
  }
}
