import "isomorphic-unfetch";
import _ from "lodash";
import http from "http";
import ws from "ws";
import fs from "fs";
import gql from "graphql-tag";
import axios from "axios";
import express from "express";
import { ApolloServer, PubSub } from "apollo-server-express";
import { ApolloClient } from "apollo-client";
import { ApolloLink } from "apollo-link";
import { BatchHttpLink as HttpLink } from "apollo-link-batch-http";
import { onError } from "apollo-link-error";
import { WebSocketLink } from "apollo-link-ws";
import { InMemoryCache } from "apollo-cache-inmemory";
import { getMainDefinition } from "apollo-utilities";
import { JsonWebToken } from "..";

const pubsub = new PubSub();

const toUser = jest.fn(_.identity);
const onSignIn = jest.fn(values => Promise.resolve(values));
const onSignOut = jest.fn(values => Promise.resolve(values));
const onRenew = jest.fn(values => Promise.resolve(values));
const resolve = jest.fn(() => 1);

const typeDefs = gql`
  type Query {
    post: Int
  }
  type Subscription {
    post: Int
  }
`;

const resolvers = {
  Query: {
    post: (payload, args, context) => resolve(context),
  },
  Subscription: {
    post: {
      subscribe: (payload, args, context) => {
        pubsub.publish("NEW_SUBSCRIPTION", context);
        return pubsub.asyncIterator(["NEW_POST"]);
      },
    },
  },
};

const app = express();

const jwt = new JsonWebToken(
  {
    private: fs.readFileSync(`${__dirname}/../../../__mocks__/private.pem`),
    public: fs.readFileSync(`${__dirname}/../../../__mocks__/public.pem`),
  },
  {
    toUser,
    onSignIn,
    onSignOut,
    onRenew,
  },
);
jwt.applyMiddleware(app);

const apolloServer = new ApolloServer({
  typeDefs,
  resolvers,
  context: jwt.contextParser(),
});
apolloServer.applyMiddleware({ app });

const server = http.createServer(app);
apolloServer.installSubscriptionHandlers(server);
server.listen();

const { port } = server.address();

const req = { token: "", correlationId: "" };
const res = {};

const httpLink = new HttpLink({
  uri: `http://localhost:${port}/graphql`,
});

const authLink = new ApolloLink((operation, forward) => {
  operation.setContext(() => ({
    headers: {
      authorization: req.token && `Bearer ${req.token}`,
    },
  }));
  return forward(operation).map(data => {
    const {
      response: { headers },
    } = operation.getContext();
    res.headers = headers;
    return data;
  });
});

const errorLink = onError(({ operation }) => {
  const {
    response: { headers },
  } = operation.getContext();
  res.headers = headers;
});

const wsLink = new WebSocketLink({
  uri: `ws://localhost:${port}/graphql`,
  options: {
    connectionParams: () => ({
      authorization: req.token && `Bearer ${req.token}`,
    }),
  },
  webSocketImpl: ws,
});

const link = ApolloLink.split(
  ({ query }) => {
    const { kind, operation } = getMainDefinition(query);
    return kind === "OperationDefinition" && operation === "subscription";
  },
  wsLink,
  ApolloLink.from([errorLink, authLink, httpLink]),
);

const client = new ApolloClient({ link, cache: new InMemoryCache() });

describe("JsonWebToken", () => {
  afterAll(() => server.close());

  describe("applyMiddleware", () => {
    it("not found method", async () => {
      await expect(
        axios.post(`http://localhost:${port}/auth`, {
          method: "notfnoud",
        }),
      ).rejects.toEqual(new Error("Request failed with status code 404"));
    });

    describe("renew token", () => {
      const correlationId = "11111111-1111-1111-1111-111111111111";

      const request = (data, id) =>
        axios.post(
          `http://localhost:${port}/auth`,
          { ...data, method: "renew" },
          { headers: { Cookie: `x-correlation-id=${id}` } },
        );

      const expectToReject = async (promise, error) => {
        let err;
        try {
          await promise;
        } catch (e) {
          err = e;
        }
        expect(err.response.data).toEqual({ error });
      };

      it("successfully renew token", async () => {
        onRenew.mockReturnValueOnce({ id: "10", isAdmin: true });
        const refreshToken = jwt.signRefresh(
          correlationId,
          { id: "10" },
          "private",
        );
        const result = await request({ refreshToken }, correlationId);
        const { data, headers } = result;
        expect(data.accessToken).not.toBeUndefined();
        expect(data.id).toBe("10");
        expect(data.isAdmin).toBe(true);
        expect(onRenew).toHaveBeenCalledTimes(1);
        expect(onRenew).toHaveBeenCalledWith({
          method: "renew",
          correlationId,
          issuer: "private",
          id: "10",
        });
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining("11111111-1111-1111-1111-111111111111"),
        );
      });

      it("successfully renew token with new correlationId", async () => {
        onRenew.mockReturnValueOnce({
          correlationId: "12345678-abcd-88cc-abcd-1111aaaa2222",
        });
        const refreshToken = jwt.signRefresh(
          correlationId,
          { id: "10" },
          "private",
        );
        const result = await request({ refreshToken }, correlationId);
        const { data, headers } = result;
        expect(data.accessToken).not.toBeUndefined();
        expect(onRenew).toHaveBeenCalledTimes(1);
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining("12345678-abcd-88cc-abcd-1111aaaa2222"),
        );
      });

      it("when reject onRenew", async () => {
        onRenew.mockRejectedValueOnce(new Error("not match to secret"));
        const refreshToken = jwt.signRefresh(
          correlationId,
          { id: "10" },
          "private",
        );
        await expectToReject(
          request({ refreshToken }, correlationId),
          "not match to secret",
        );
      });

      it("when not match to correlation", async () => {
        const refreshToken = jwt.signRefresh(
          correlationId,
          { id: "10" },
          "private",
        );
        await expectToReject(
          request({ refreshToken }, "12345678-abcd-88cc-abcd-1111aaaa2222"),
          "not match to correlation id",
        );
      });
    });

    describe("signIn", () => {
      const correlationId = "11111111-1111-1111-1111-111111111111";

      const request = (data, id) =>
        axios.post(
          `http://localhost:${port}/auth`,
          { ...data, method: "signIn" },
          { headers: { Cookie: `x-correlation-id=${id || ""}` } },
        );

      it("successfully signIn", async () => {
        onSignIn.mockReturnValueOnce({
          correlationId,
          id: "10",
          issuer: "private",
        });
        const body = { username: "apple", password: "applepen" };
        const result = await request(body);
        const { data, headers } = result;
        expect(data.refreshToken).not.toBeUndefined();
        expect(data.accessToken).not.toBeUndefined();
        expect(data.id).toBe("10");
        expect(onSignIn).toHaveBeenCalledTimes(1);
        expect(onSignIn).toHaveBeenCalledWith({
          method: "signIn",
          correlationId: "",
          ...body,
        });
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining("11111111-1111-1111-1111-111111111111"),
        );
      });

      it("successfully signIn with no return correlationId", async () => {
        onSignIn.mockReturnValueOnce({ id: "10", issuer: "private" });
        const body = { username: "banana", password: "bananapen" };
        const result = await request(body);
        const { data, headers } = result;
        expect(data.refreshToken).not.toBeUndefined();
        expect(data.accessToken).not.toBeUndefined();
        expect(data.id).toBe("10");
        expect(onSignIn).toHaveBeenCalledTimes(1);
        expect(onSignIn).toHaveBeenCalledWith({
          method: "signIn",
          correlationId: "",
          ...body,
        });
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining("x-correlation-id="),
        );
      });

      it("successfully signIn with already have correlationId", async () => {
        onSignIn.mockReturnValueOnce({ id: "10", issuer: "private" });
        const body = { username: "banana", password: "bananapen" };
        const result = await request(body, correlationId);
        const { data, headers } = result;
        expect(data.refreshToken).not.toBeUndefined();
        expect(data.accessToken).not.toBeUndefined();
        expect(data.id).toBe("10");
        expect(onSignIn).toHaveBeenCalledTimes(1);
        expect(onSignIn).toHaveBeenCalledWith({
          method: "signIn",
          correlationId,
          ...body,
        });
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining(correlationId),
        );
      });

      it("when isUnfamiliar", async () => {
        onSignIn.mockReturnValueOnce({
          correlationId,
          isUnfamiliar: true,
          id: "10",
        });
        const body = { username: "apple", password: "applepen" };
        const result = await request(body);
        const { data, headers } = result;
        expect(data.refreshToken).toBeUndefined();
        expect(data.accessToken).toBeUndefined();
        expect(data.id).toBe("10");
        expect(data.isUnfamiliar).toBe(true);
        expect(onSignIn).toHaveBeenCalledTimes(1);
        expect(onSignIn).toHaveBeenCalledWith({
          method: "signIn",
          correlationId: "",
          ...body,
        });
        expect(headers["set-cookie"][0]).toEqual(
          expect.stringContaining("11111111-1111-1111-1111-111111111111"),
        );
      });

      it("when single key", async () => {
        const signature = new JsonWebToken({
          private: fs.readFileSync(
            `${__dirname}/../../../__mocks__/private.pem`,
          ),
        });
        const result = await signature.signIn({});
        expect(result.refreshToken).not.toBeUndefined();
        expect(result.accessToken).not.toBeUndefined();
      });
    });

    describe("signOut", () => {
      const correlationId = "11111111-1111-1111-1111-111111111111";

      const request = (data, id) =>
        axios.post(
          `http://localhost:${port}/auth`,
          { ...data, method: "signOut" },
          { headers: { Cookie: `x-correlation-id=${id || ""}` } },
        );

      it("successfully signOut", async () => {
        onSignOut.mockReturnValueOnce({});
        const result = await request({}, correlationId);
        const { data, headers } = result;
        expect(data.status).toBe("ok");
        expect(onSignOut).toHaveBeenCalledTimes(1);
        expect(onSignOut).toHaveBeenCalledWith({
          method: "signOut",
          correlationId,
        });
        expect(headers["set-cookie"]).toBeUndefined();
      });
    });
  });

  describe("contextParser", () => {
    describe("http", () => {
      const request = (token = "") => {
        req.token = token;
        return client.query({
          query: gql`
            query {
              post
            }
          `,
          fetchPolicy: "network-only",
        });
      };

      const expectToReject = async (query, contentEextend) => {
        await expect(
          query.catch(error => {
            const { errors } = error.networkError.result;
            expect(errors[0].message).toBe(
              "Context creation failed: token invalid",
            );
            throw error;
          }),
        ).rejects.toEqual(
          new Error(
            "Network error: Response not successful: Received status code 400",
          ),
        );
        expect(toUser).toHaveBeenCalledTimes(0);
        expect(res.headers.get("x-content-extend")).toBe(contentEextend);
      };

      it("successfully get guest", async () => {
        const { data } = await request();
        expect(data).toEqual({ post: 1 });
        expect(resolve).not.toHaveBeenCalledWith(
          expect.objectContaining({ user: expect.anything() }),
        );
        expect(resolve).toHaveBeenCalledTimes(1);
      });

      it("successfully get user", async () => {
        const user = { id: "1" };
        const token = jwt.signAccess(user, "private");
        const { data } = await request(token);

        expect(data).toEqual({ post: 1 });
        expect(toUser).toHaveBeenCalledTimes(1);

        expect(resolve).toHaveBeenCalledWith(
          expect.objectContaining({ user: { id: "1" }, viewer: { id: "1" } }),
        );
      });

      it("when token expired", async () => {
        const token = jwt.sign({ user: { id: "10" }, exp: 0 }, "private");
        await expectToReject(request(token), "expired");
      });

      it("when token is invalid", async () => {
        await expectToReject(request("XXYYZZ"), "decode malformed");
      });

      it("when use refreshToken", async () => {
        const correlationId = "11111111-1111-1111-1111-111111111111";
        const token = jwt.signRefresh(correlationId, { id: "10" }, "private");
        await expectToReject(request(token), "subject invalid");
      });
    });

    describe("subscription", () => {
      const subscribe = async (token = "") => {
        req.token = token;
        wsLink.subscriptionClient.close();
        let subId;
        const result = await new Promise(async (subResolve, subReject) => {
          subId = await pubsub.subscribe("NEW_SUBSCRIPTION", subResolve);
          client
            .subscribe({
              query: gql`
                subscription {
                  post
                }
              `,
            })
            .subscribe({ error: subReject });
        });
        pubsub.unsubscribe(subId);
        return result;
      };

      it("successfully get guest", async () => {
        const result = await subscribe("");
        expect(result).not.toEqual(
          expect.objectContaining({ user: expect.anything() }),
        );
      });

      it("successfully get user", async () => {
        const user = { id: "1" };
        const token = jwt.signAccess(user, "private");
        const result = await subscribe(token);
        expect(result).toEqual(
          expect.objectContaining({ user: { id: "1" }, viewer: { id: "1" } }),
        );
      });

      it("when token expired", async () => {
        const token = jwt.sign({ id: "10", exp: 0 }, "private");
        await expect(subscribe(token)).rejects.toEqual({
          message: "expired",
        });
      });

      it("when token is invalid", async () => {
        const token = "XXYYZZ";
        await expect(subscribe(token)).rejects.toEqual({
          message: "decode malformed",
        });
      });
    });
  });
});
